/*
 * xvgif.c  -  GIF loading code for 'xv'.  Based strongly on...
 *
 * gif2ras.c - Converts from a Compuserve GIF (tm) image to a Sun Raster image.
 *
 * Copyright (c) 1988, 1989 by Patrick J. Naughton
 *
 * Author: Patrick J. Naughton
 * naughton@wind.sun.com
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 */

/* Copyright Notice
 * ================
 * Copyright 1989, 1990, 1991, 1992, 1993 by John Bradley
 * 
 *
 *
 * This file was heavily modified by Khramov Leo (old@sunct2.jinr.dubna.su)
 * for using with ImageEngine and for speeding up GIF loading.....
 * (c) Leo 1996.
 *
 *
 ***********************************************************/
 
#define JPEG_IMAGES
#include "picinfo.h"
#include "accel.h"
#include "image.h"

extern int bdep, b_pad;

typedef int boolean;

#define NEXTBYTE (*dataptr++)
#define EXTENSION     0x21
#define IMAGESEP      0x2c
#define TRAILER       0x3b
#define INTERLACEMASK 0x40
#define COLORMAPMASK  0x80



int numcols;
float normaspect;

extern "C" void xvbzero(char* s,int len);


int BitOffset = 0,              /* Bit Offset of next code */
    XC = 0, YC = 0,             /* Output X and Y coords of current pixel */
    Pass = -1,                  /* Used by output routine if interlaced pic */
    OutCount = 0,               /* Decompressor output 'stack count' */
    RWidth, RHeight,            /* screen dimensions */
    Width, Height,              /* image dimensions */
    L_Width = 0,                /* Line-width including pading */
    LeftOfs, TopOfs,            /* image offset */
    BitsPerPixel,               /* Bits per pixel, read from GIF header */
    BytesPerScanline,           /* bytes per scanline in output raster */
    ColorMapSize,               /* number of colors */
    Background,                 /* background color */
    CodeSize,                   /* Code size, read from GIF header */
    InitCodeSize,               /* Starting code size, used during Clear */
    Code,                       /* Value returned by ReadCode */
    MaxCode,                    /* limiting value for current code size */
    ClearCode,                  /* GIF clear code */
    EOFCode,                    /* GIF end-of-information code */
    CurCode, OldCode, InCode,   /* Decompressor variables */
    FirstFree,                  /* First free code, generated per GIF spec */
    FreeCode,                   /* Decompressor,next free slot in hash table */
    FinChar,                    /* Decompressor variable */
    BitMask,                    /* AND mask for data size */
    ReadMask,                   /* Code AND mask for current code size */
    Misc;                       /* miscellaneous bits (interlace, local cmap)*/


boolean Interlace, HasColormap;

byte *RawGIF;                   /* The heap array to hold it, raw */
byte *Raster;                   /* The raster data stream, unblocked */
byte *pic8;

    /* The hash table used by the decompressor */
int Prefix[4096];
int Suffix[4096];

    /* An output array used by the decompressor */
int OutCode[4097];

int   gif89 = 0;
char *id87 = "GIF87a";
char *id89 = "GIF89a";

static int EGApalette[16][3] = {
  {0,0,0},       {0,0,128},     {0,128,0},     {0,128,128},
  {128,0,0},     {128,0,128},   {128,128,0},   {200,200,200},
  {100,100,100}, {100,100,255}, {100,255,100}, {100,255,255},
  {255,100,100}, {255,100,255}, {255,255,100}, {255,255,255} };


int   readImage(PICINFO*);
int   ReadCode();
void  DoInterlace(byte);
int   gifError(PICINFO*,char*);
void  gifWarning(char*);

int   filesize;
char *bname=NULL;

byte *dataptr;
extern int fromMem;

int deGIF(char* fbu,PICINFO* pinfo,int fsiz);

/*****************************/
int LoadGIF(char* fname,PICINFO* pinfo)
/*****************************/
{
  /* returns '1' if successful */

  pinfo->pic     = (byte *) NULL;

  bname =fname;
  int fpp=open(fname,O_RDONLY);
  if (fpp==-1) return ( gifError(pinfo, "can't open file") );

  filesize=lseek(fpp, 0L, 2);
  lseek(fpp, 0L, 0);

  /* the +256's are so we can read truncated GIF files without fear of
     segmentation violation */
  if (!(dataptr = RawGIF = (byte *) calloc(filesize+256,1)))
    return( gifError(pinfo, "not enough memory to read gif file") );

  if (read(fpp,dataptr, filesize) != filesize)
    return( gifError(pinfo, "GIF data read failed") );
  close(fpp);
  return deGIF((char*) dataptr,pinfo,filesize);
}

/*****************************/
int deGIF(char* fbu,PICINFO* pinfo,int fsiz)
/*****************************/
{
  /* returns '1' if successful */

  register byte  ch, *origptr;
  register int   i, block;
  int            aspect, gotimage;

  /* initialize variables */
  BitOffset = XC = YC = OutCount = gotimage = 0;
  Pass = -1;
  RawGIF = Raster = pic8 = NULL;
  gif89 = 0;

  pinfo->pic     = (byte *) NULL;
  pinfo->comment = (char *) NULL;

  /* find the size of the file */
  filesize=fsiz;

  /* the +256's are so we can read truncated GIF files without fear of
     segmentation violation */
//  if (!(Raster = (byte *) calloc(filesize+256,1)))
//    return( gifError(pinfo, "not enough memory to read gif file") );

  dataptr=Raster=RawGIF=(byte*)fbu;

  origptr = dataptr;

  if      (strncmp((char *) dataptr, id87, 6)==0) gif89 = 0;
  else if (strncmp((char *) dataptr, id89, 6)==0) gif89 = 1;
  else    return( gifError(pinfo, "not a GIF file"));

  dataptr += 6;

  /* Get variables from the GIF screen descriptor */

  ch = NEXTBYTE;
  RWidth = ch + 0x100 * NEXTBYTE;       /* screen dimensions... not used. */
  ch = NEXTBYTE;
  RHeight = ch + 0x100 * NEXTBYTE;

  ch = NEXTBYTE;
  HasColormap = ((ch & COLORMAPMASK) ? True : False);

  BitsPerPixel = (ch & 7) + 1;
  numcols = ColorMapSize = 1 << BitsPerPixel;
  BitMask = ColorMapSize - 1;

  Background = NEXTBYTE;                /* background color... not used. */

  aspect = NEXTBYTE;
  if (aspect) {
    if (!gif89) return(gifError(pinfo,"corrupt GIF file (screen descriptor)"));
    else normaspect = (float) (aspect + 15) / 64.0;   /* gif89 aspect ratio */
  }


  /* Read in global colormap. */

  if (HasColormap)
    for (i=0; i<ColorMapSize; i++) {
      pinfo->pal[i*3] = NEXTBYTE;
      pinfo->pal[i*3+1] = NEXTBYTE;
      pinfo->pal[i*3+2] = NEXTBYTE;
    }
  else {  /* no colormap in GIF file */
    /* put std EGA palette (repeated 16 times) into colormap, for lack of
       anything better to do */

    for (i=0; i<256; i++) {
      pinfo->pal[i*3] = EGApalette[i&15][0];
      pinfo->pal[i*3+1] = EGApalette[i&15][1];
      pinfo->pal[i*3+2] = EGApalette[i&15][2];
    }
  }

  /* possible things at this point are:
   *   an application extension block
   *   a comment extension block
   *   an (optional) graphic control extension block
   *       followed by either an image
   *       or a plaintext extension
   */

  while (dataptr-RawGIF<filesize) {
    block = NEXTBYTE;


    if (block == EXTENSION) {  /* parse extension blocks */
      int i, fn, blocksize, aspnum, aspden;

      /* read extension block */
      fn = NEXTBYTE;


      if (fn == 'R') {                  /* GIF87 aspect extension */
        int sbsize;

        blocksize = NEXTBYTE;
        if (blocksize == 2) {
          aspnum = NEXTBYTE;
          aspden = NEXTBYTE;
          if (aspden>0 && aspnum>0)
            normaspect = (float) aspnum / (float) aspden;
          else { normaspect = 1.0;  aspnum = aspden = 1; }

        }
        else {
          for (i=0; i<blocksize; i++) NEXTBYTE;
        }

        while ((sbsize=NEXTBYTE)>0) {  /* eat any following data subblocks */
          for (i=0; i<sbsize; i++) NEXTBYTE;
        }
      }


      else if (fn == 0xFE) {  /* Comment Extension */
        int   ch, j, sbsize, cmtlen;
        byte *ptr1, *cmt, *cmt1, *sp;

        cmtlen = 0;
        ptr1 = dataptr;      /* remember start of comments */

        /* figure out length of comment */
        do {
          sbsize = NEXTBYTE;
          cmtlen += sbsize;
          for (j=0; j<sbsize; j++) ch = NEXTBYTE;
        } while (sbsize);


        if (cmtlen>0) {   /* build into one un-blocked comment */
          cmt = (byte *) malloc(cmtlen);
          if (!cmt) gifWarning("couldn't malloc space for comments\n");
          else {
            sp = cmt;
            do {
              sbsize = (*ptr1++);
              for (j=0; j<sbsize; j++, sp++, ptr1++) *sp = *ptr1;
            } while (sbsize);

            if (pinfo->comment) {    /* have to strcat onto old comments */
              cmt1 = (byte *) malloc(strlen(pinfo->comment) + cmtlen);
              if (!cmt1) {
                gifWarning("couldn't malloc space for comments\n");
                free(cmt);
              }
              else {
                strcpy((char *) cmt1, (char *) pinfo->comment);
                strcat((char *) cmt1, (char *) cmt);
                free(pinfo->comment);
                free(cmt);
                pinfo->comment = (char *) cmt1;
              }
            }
            else pinfo->comment = (char *) cmt;
          }  /* if (cmt) */
        }  /* if cmtlen>0 */
      }


      else if (fn == 0x01) {  /* PlainText Extension */
        int j,sbsize,ch;
        int tgLeft, tgTop, tgWidth, tgHeight, cWidth, cHeight, fg, bg;

        fprintf(stderr, "%s:  %s", bname,
                "PlainText extension found in GIF file.  Ignored.");

        sbsize   = NEXTBYTE;
        tgLeft   = NEXTBYTE;  tgLeft   += (NEXTBYTE)<<8;
        tgTop    = NEXTBYTE;  tgTop    += (NEXTBYTE)<<8;
        tgWidth  = NEXTBYTE;  tgWidth  += (NEXTBYTE)<<8;
        tgHeight = NEXTBYTE;  tgHeight += (NEXTBYTE)<<8;
        cWidth   = NEXTBYTE;
        cHeight  = NEXTBYTE;
        fg       = NEXTBYTE;
        bg       = NEXTBYTE;
        i=12;
        for ( ; i<sbsize; i++) NEXTBYTE;   /* read rest of first subblock */

        /* read (and ignore) data sub-blocks */
        do {
          j = 0;
          sbsize = NEXTBYTE;
          while (j<sbsize) {
            ch = NEXTBYTE;  j++;
          }
        } while (sbsize);
      }


      else if (fn == 0xF9) {  /* Graphic Control Extension */
        int j, sbsize;

        /* read (and ignore) data sub-blocks */
        do {
          j = 0; sbsize = NEXTBYTE;
          while (j<sbsize) { NEXTBYTE;  j++; }
        } while (sbsize);
      }


      else if (fn == 0xFF) {  /* Application Extension */
        int j, sbsize;


        /* read (and ignore) data sub-blocks */
        do {
          j = 0; sbsize = NEXTBYTE;
          while (j<sbsize) { NEXTBYTE;  j++; }
        } while (sbsize);
      }


      else { /* unknown extension */
        int j, sbsize;


        /* read (and ignore) data sub-blocks */
        do {
          j = 0; sbsize = NEXTBYTE;
          while (j<sbsize) { NEXTBYTE;  j++; }
        } while (sbsize);
      }
    }


    else if (block == IMAGESEP) {

      if (gotimage) {   /* just skip over remaining images */
        int i,misc,ch,ch1;

        /* skip image header */
        NEXTBYTE;  NEXTBYTE;  /* left position */
        NEXTBYTE;  NEXTBYTE;  /* top position */
        NEXTBYTE;  NEXTBYTE;  /* width */
        NEXTBYTE;  NEXTBYTE;  /* height */
        misc = NEXTBYTE;      /* misc. bits */

        if (misc & 0x80) {    /* image has local colormap.  skip it */
          for (i=0; i< 1 << ((misc&7)+1);  i++) {
            NEXTBYTE;  NEXTBYTE;  NEXTBYTE;
          }
        }

        NEXTBYTE;       /* minimum code size */

        /* skip image data sub-blocks */
        do {
          ch = ch1 = NEXTBYTE;
          while (ch--) NEXTBYTE;
          if ((dataptr - RawGIF) > filesize) break;      /* EOF */
        } while(ch1);
      }

      else if (readImage(pinfo)) gotimage = 1;
    }


    else if (block == TRAILER) {      /* stop reading blocks */
      break;
    }

    else {      /* unknown block type */
      char str[128];


      /* don't mention bad block if file was trunc'd, as it's all bogus */
      if ((dataptr - origptr) < filesize) {
        sprintf(str, "Unknown block type (0x%02x) at offset 0x%x",
                block, (dataptr - origptr) - 1);

        if (!gotimage) return gifError(pinfo, str);
        else gifWarning(str);
      }

      break;
    }

  }

  if(fromMem==0)
     free(RawGIF);  
  RawGIF = NULL;
//  free(Raster);  Raster = NULL;

  if (!gotimage)
     return( gifError(pinfo, "no image data found in GIF file") );

  return 1;
}

int leo_blen=0;
int leo_coun=0;

/********************************************/
int readImage(PICINFO* pinfo)
{
  register byte ch, *picptr;
  int           i, npixels, maxpixels;

  npixels = maxpixels = 0;

  /* read in values from the image descriptor */

  ch = NEXTBYTE;
  LeftOfs = ch + 0x100 * NEXTBYTE;
  ch = NEXTBYTE;
  TopOfs  = ch + 0x100 * NEXTBYTE;
  ch = NEXTBYTE;
  Width   = ch + 0x100 * NEXTBYTE;

  ch = NEXTBYTE;
  Height  = ch + 0x100 * NEXTBYTE;

  Misc = NEXTBYTE;
  Interlace = ((Misc & INTERLACEMASK) ? True : False);
  if (Interlace)
     Pass = 0;
  if (Misc & 0x80) {
    for (i=0; i< 1 << ((Misc&7)+1); i++) {
      pinfo->pal[i*3] = NEXTBYTE;
      pinfo->pal[i*3+1] = NEXTBYTE;
      pinfo->pal[i*3+2] = NEXTBYTE;
    }
  }


  if (!HasColormap && !(Misc&0x80)) {
    /* no global or local colormap */
//    fprintf(stderr, "%s:  %s", bname,
//            "No colormap in this GIF file.  Assuming EGA colors.");
  }

   im_init_translation();

  if (translation) {
     L_Width = Width * bdep;
     if (L_Width % b_pad) {
        L_Width += b_pad - L_Width % b_pad;
     }
  } else {
     L_Width = Width;
  }
  /* Start reading the raster data. First we get the intial code size
   * and compute decompressor constant values, based on this code size.
   */

  CodeSize = NEXTBYTE;

  ClearCode = (1 << CodeSize);
  EOFCode = ClearCode + 1;
  FreeCode = FirstFree = ClearCode + 2;

  /* The GIF spec has it that the code size is the code size used to
   * compute the above values is the code size given in the file, but the
   * code size used in compression/decompression is the code size given in
   * the file plus one. (thus the ++).
   */

  CodeSize++;
  InitCodeSize = CodeSize;
  MaxCode = (1 << CodeSize);
  ReadMask = MaxCode - 1;



 leo_blen=leo_coun=0;
 Raster=dataptr;

  /* Allocate the 'pic' */
  maxpixels = Width*Height;
if(translation==0)
  picptr = pic8 = (byte *) malloc(maxpixels);
 else //{
  picptr = pic8 = (byte *) malloc(Height * L_Width);
  if (!pic8) return( gifError(pinfo, "couldn't malloc 'pic8'") );
      

  /* Decompress the file, continuing until you see the GIF EOF code.
   * One obvious enhancement is to add checking for corrupt files here.
   */

  Code = ReadCode();
  while (Code != EOFCode) {
    /* Clear code sets everything back to its initial value, then reads the
     * immediately subsequent code as uncompressed data.
     */

    if (Code == ClearCode) {
      CodeSize = InitCodeSize;
      MaxCode = (1 << CodeSize);
      ReadMask = MaxCode - 1;
      FreeCode = FirstFree;
      Code = ReadCode();
      CurCode = OldCode = Code;
      FinChar = CurCode & BitMask;
#if 0
      if (!Interlace) im_coding256((char**)&picptr,FinChar); //*picptr++ = FinChar;
         else DoInterlace(FinChar);
#else
      DoInterlace(FinChar);
#endif
      npixels++;
    }
    else {
      /* If not a clear code, must be data: save same as CurCode and InCode */

      /* if we're at maxcode and didn't get a clear, stop loading */
      if (FreeCode>=4096) { /* printf("freecode blew up\n"); */
                            break; }

      CurCode = InCode = Code;

      /* If greater or equal to FreeCode, not in the hash table yet;
       * repeat the last character decoded
       */

      if (CurCode >= FreeCode) {
        CurCode = OldCode;
        if (OutCount > 4096) {  /* printf("outcount1 blew up\n"); */ break; }
        OutCode[OutCount++] = FinChar;
      }

      /* Unless this code is raw data, pursue the chain pointed to by CurCode
       * through the hash table to its end; each code in the chain puts its
       * associated output code on the output queue.
       */

      while (CurCode > BitMask) {
        if (OutCount > 4096) break;   /* corrupt file */
        OutCode[OutCount++] = Suffix[CurCode];
        CurCode = Prefix[CurCode];
      }

      if (OutCount > 4096) { /* printf("outcount blew up\n"); */ break; }

      /* The last code in the chain is treated as raw data. */

      FinChar = CurCode & BitMask;
      OutCode[OutCount++] = FinChar;

      /* Now we put the data out to the Output routine.
       * It's been stacked LIFO, so deal with it that way...
       */

      /* safety thing:  prevent exceeding range of 'pic8' */
      if (npixels + OutCount > maxpixels) OutCount = maxpixels-npixels;

      npixels += OutCount;
#if 0
      if (!Interlace) for (i=OutCount-1; i>=0; i--) im_coding256((char**)&picptr,OutCode[i]);//*picptr++ = OutCode[i];
                else  for (i=OutCount-1; i>=0; i--) DoInterlace(OutCode[i]);
#else
      for (i=OutCount-1; i>=0; i--) DoInterlace(OutCode[i]);
#endif
      OutCount = 0;

      /* Build the hash table on-the-fly. No table is stored in the file. */
                               
      Prefix[FreeCode] = OldCode;
      Suffix[FreeCode] = FinChar;
      OldCode = InCode;

      /* Point to the next slot in the table.  If we exceed the current
       * MaxCode value, increment the code size unless it's already 12.  If it
       * is, do nothing: the next code decompressed better be CLEAR
       */

      FreeCode++;
      if (FreeCode >= MaxCode) {
        if (CodeSize < 12) {
          CodeSize++;
          MaxCode *= 2;
          ReadMask = (1 << CodeSize) - 1;
        }
      }
    }
    Code = ReadCode();
    if (npixels >= maxpixels) break;
  }

  if (npixels != maxpixels) {
//    fprintf(stderr,"%s:  %s", bname,
//            "This GIF file seems to be truncated.  Winging it.");
    if (!Interlace)
      xvbzero((char*)pic8+npixels, maxpixels-npixels);  /* clear to EOBuffer */
  }

//  fclose(fp);

  /* fill in the PICINFO structure */

  pinfo->pic     = pic8;
  pinfo->w       = Width;
  pinfo->h       = Height;
  pinfo->type    = PIC8;
  pinfo->frmType = F_GIF;
  pinfo->colType = F_FULLCOLOR;

  sprintf(pinfo->fullInfo,
          "%dx%d GIF%s, %d bits per pixel, %sinterlaced.  (%d bytes)",
          Width,Height,(gif89) ? "89" : "87", BitsPerPixel,
          Interlace ? "" : "non-", filesize);

  sprintf(pinfo->shrtInfo, "%dx%d GIF%s.",Width,Height,(gif89) ? "89" : "87");

  /* pinfo.comment gets handled in main LoadGIF() block-reader */
  dataptr+=filesize;
  return 1;
}



/* Fetch the next code from the raster data stream.  The codes can be
 * any length from 3 to 12 bits, packed into 8-bit bytes, so we have to
 * maintain our location in the Raster array as a BIT Offset.  We compute
 * the byte Offset into the raster array by dividing this by 8, pick up
 * three bytes, compute the bit Offset into our 24-bit chunk, shift to
 * bring the desired code to the bottom, then mask it off and return it.
 */

int ReadCode()
{
  int RawCode, ByteOffset, leo_bplus,leo_bbplus;

  ByteOffset = BitOffset >> 3;
  if(ByteOffset>=leo_blen)
   {
     Raster+=leo_blen;
     ByteOffset-=leo_blen;
     BitOffset-=(leo_blen<<3);
     leo_blen=(*Raster);
     Raster++;
    };
   leo_bplus=ByteOffset+1;
   leo_bbplus=leo_bplus+1;
   if(leo_bplus>=leo_blen) leo_bplus++;
   if(leo_bbplus>=leo_blen) leo_bbplus++;
  RawCode = Raster[ByteOffset] + ((int)(Raster[leo_bplus]) << 8);
  if (CodeSize >= 8)
    RawCode += ( ((int) Raster[leo_bbplus]) << 16);
  RawCode >>= (BitOffset & 7);
  BitOffset += CodeSize;

  return(RawCode & ReadMask);
}

/***************************/
void DoInterlace(byte Index)
{
  static byte *ptr = NULL;
  static int   oldYC = -1;

  if (oldYC != YC) {  ptr = pic8 + YC * L_Width;  oldYC = YC; }

  if (YC < Height)
    im_coding256((char**)&ptr,Index);//*ptr++ = Index;

  /* Update the X-coordinate, and if it overflows, update the Y-coordinate */

  if (++XC == Width) {

        
     /* deal with the interlace as described in the GIF
      * spec.  Put the decoded scan line out to the screen if we haven't gone
      * past the bottom of it
      */
     
     XC = 0;
     
     switch (Pass) {
     case 0:
        YC += 8;
        if (YC >= Height) { Pass++; YC = 4; }
        break;
        
     case 1:
        YC += 8;
        if (YC >= Height) { Pass++; YC = 2; }
        break;
        
     case 2:
        YC += 4;
        if (YC >= Height) { Pass++; YC = 1; }
        break;
        
     case 3:
        YC += 2;  break;
        
     default:
        YC++;
        break;
     }
     
  }
}



/*****************************/
int gifError(PICINFO* pinfo,char* st)
{
  gifWarning(st);

  if (RawGIF != NULL && fromMem==0) free(RawGIF);
//  if (Raster != NULL) free(Raster);

  if (pinfo->pic) free(pinfo->pic);
  if (pinfo->comment) free(pinfo->comment);

  if (pic8 && pic8 != pinfo->pic) free(pic8);

  pinfo->pic = (byte *) NULL;
  pinfo->comment = (char *) NULL;

  return 0;
}


/*****************************/
void gifWarning(char* st)
{
 if(bname)
  fprintf(stderr,"%s:  %s\n", bname, st);
 else
       fprintf(stderr,"GIF: %s\n",st);
}
